use std::collections::HashMap;

use proc_macro2::{Ident, TokenStream};
use quote::{format_ident, quote, ToTokens};

use crate::{
    case_convert::CaseConvert,
    model::{
        ActorKind, ActorModel, MethodModel, MsgInfo, OutputKind, ProtocolModel, TypeParamInfo,
    },
};

impl ToTokens for ProtocolModel {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        tokens.extend(self.generate_message_enum());
        tokens.extend(self.generate_state_traits());
        tokens.extend(self.generate_spawn_functions());
    }
}

impl ProtocolModel {
    fn generate_message_enum(&self) -> TokenStream {
        let msg_lookup = self.msg_lookup();
        let variants: Vec<_> = msg_lookup.msg_iter().map(|msg| msg.msg_name()).collect();
        let variant_names_map: HashMap<&Ident, String> = variants
            .iter()
            .map(|variant| (variant.as_ref(), variant.to_string()))
            .collect();
        let msg_types = msg_lookup.msg_iter().map(|msg| msg.msg_type());
        let all_replies = msg_lookup.msg_iter().all(|msg| msg.is_reply());
        let enum_name = self.msg_enum_ident();
        let enum_kinds_name = self.msg_enum_kinds_ident();
        let name_decl_vec: Vec<(Ident, TokenStream)> = variants
            .iter()
            .map(|variant| {
                let variant_str = variant_names_map.get(variant.as_ref()).unwrap();
                let name_ident =
                    format_ident!("{}_NAME", variant_str.pascal_to_snake().to_uppercase());
                let decl = quote! {
                    static #name_ident: Lazy<Arc<String>> = Lazy::new(|| {
                        Arc::new(#variant_str.into())
                    });
                };
                (name_ident, decl)
            })
            .collect();
        let name_decls = name_decl_vec.iter().map(|(_, decl)| decl);
        let name_idents = name_decl_vec.iter().map(|(ident, _)| ident);
        let send_impl = if all_replies {
            quote! {}
        } else {
            quote! {
                impl ::btrun::model::SendMsg for #enum_name {}
            }
        };
        let proto_name = &self.def().name_def.name;
        let doc_comment = format!("Message type for the {proto_name} protocol.");
        quote! {
            #[doc = #doc_comment]
            #[derive(::serde::Serialize, ::serde::Deserialize)]
            pub enum #enum_name {
                #( #variants(#msg_types) ),*
            }


            #[derive(Clone)]
            #[allow(dead_code)]
            pub enum #enum_kinds_name {
                #( #variants ),*
            }

            impl ::btrun::model::Named for #enum_kinds_name {
                fn name(&self) -> ::std::sync::Arc<String> {
                    use ::btrun::model::Lazy;
                    use ::std::sync::Arc;
                    #( #name_decls )*
                    match self {
                        #( Self::#variants => #name_idents.clone() ),*
                    }
                }
            }

            impl Copy for #enum_kinds_name {}

            impl ::btrun::model::Named for #enum_name {
                fn name(&self) -> ::std::sync::Arc<String> {
                    match self {
                        #( Self::#variants(_) => #enum_kinds_name::#variants.name()),*
                    }
                }
            }

            impl ::btrun::model::CallMsg for #enum_name {
                type Reply = Self;
            }

            #send_impl
        }
    }

    fn generate_state_traits(&self) -> TokenStream {
        let traits = self.states_iter().map(|state| {
            (
                state.name(),
                state.methods().values(),
                self.actor_lookup()
                    .actor_with_init_state(state.name())
                    .is_some(),
            )
        });
        let mut tokens = TokenStream::new();
        for (trait_ident, methods, is_init_state) in traits {
            let method_tokens = methods.map(|x| x.generate_tokens());
            let actor_impl_method = if is_init_state {
                quote! {
                    #[doc = "The name of the implementation for the actor this state is a part of."]
                    fn actor_impl() -> ::std::sync::Arc<String>;
                }
            } else {
                quote! {}
            };
            let trait_str = format!("{trait_ident}");
            quote! {
                pub trait #trait_ident : Send + Sync + Sized {
                    #actor_impl_method
                    #( #method_tokens )*

                    #[doc = "The name of this state."]
                    fn state_name() -> ::std::sync::Arc<String> {
                        use ::btrun::model::Lazy;
                        use ::std::sync::Arc;
                        static STATE_NAME: Lazy<Arc<String>> = Lazy::new(|| {
                            Arc::new(#trait_str.into())
                        });
                        STATE_NAME.clone()
                    }
                }
            }
            .to_tokens(&mut tokens);
        }
        tokens
    }

    fn generate_spawn_functions(&self) -> TokenStream {
        let mut output = TokenStream::default();
        for actor in self.actors_iter() {
            output.extend(self.generate_spawn_function(actor));
        }
        output
    }

    fn generate_spawn_function(&self, actor: &ActorModel) -> TokenStream {
        match actor.kind() {
            ActorKind::Service => self.generate_service_spawn_function(actor),
            ActorKind::Worker => self.generate_worker_spawn_function(actor),
            ActorKind::Client => TokenStream::default(),
        }
    }

    fn generate_worker_spawn_function(&self, actor: &ActorModel) -> TokenStream {
        let function_name = actor.spawn_function_ident();
        let TypeParamInfo {
            type_params,
            constraints,
        } = self.type_param_info_for(actor.name());
        let init_state_type_param = actor.init_state().type_param();
        let state_enum_decl = self.generate_state_enum(actor);
        let init_state_var = self.init_state_var();
        let server_loop = self.generate_loop(actor);
        quote! {
            async fn #function_name<#( #type_params ),*>(
                runtime: &'static ::btrun::Runtime,
                owner_name: ::btrun::model::ActorName,
                #init_state_var: #init_state_type_param,
            ) -> ::btlib::Result<::btrun::model::ActorName>
            where
                #( #constraints ),*
            {
                #state_enum_decl

                let actor = #server_loop ;
                runtime.spawn(None, actor).await
            }
        }
    }

    fn generate_service_spawn_function(&self, actor: &ActorModel) -> TokenStream {
        let function_name = format_ident!("register_{}", actor.def().actor.as_ref());
        let TypeParamInfo {
            type_params,
            constraints,
        } = self.type_param_info_for(actor.name());
        let init_state_type_param = actor.init_state().type_param();
        let state_enum_decl = self.generate_state_enum(actor);
        let msg_enum = self.msg_enum_ident();
        let init_state_var = self.init_state_var();
        let server_loop = self.generate_loop(actor);
        quote! {
            async fn #function_name<#( #type_params ),*, F>(
                runtime: &'static ::btrun::Runtime,
                service_id: ::btrun::model::ServiceId,
                make_init: F,
            ) -> ::btlib::Result<::btrun::model::ServiceName>
            where
                F: 'static + Send + Sync + Clone + Fn() -> #init_state_type_param,
                #( #constraints ),*
            {
                #state_enum_decl

                runtime.register::<#msg_enum, _>(
                    service_id,
                    move |runtime: &'static ::btrun::Runtime| {
                        let make_init = make_init.clone();
                        let fut = async move {
                            let #init_state_var = make_init();
                            let server_loop = #server_loop ;
                            // This shouldn't panic because we passed in None as the owner.
                            let actor_impl = runtime.spawn(None, server_loop).await.unwrap();
                            Ok(actor_impl)
                        };
                        ::std::boxed::Box::pin(fut)
                    }
                ).await
            }
        }
    }

    fn generate_loop(&self, actor: &ActorModel) -> TokenStream {
        let init_state = actor.init_state().name();
        let state_enum_ident = actor.state_enum_ident();
        let msg_enum = self.msg_enum_ident();
        let init_state_var = self.init_state_var();
        let end_ident = self.end_ident();
        let mailbox = self.mailbox_param();
        let actor_id = self.actor_id_param();
        let runtime = self.runtime_param();
        let from = self.from_ident();
        let reply = self.reply_ident();
        let msg = self.msg_ident();
        let actor_name = self.actor_name_ident();
        let call_transitions = self.generate_call_transitions(actor);
        let send_transitions = self.generate_send_transitions(actor);
        let control_transitions = self.generate_control_transitions(actor);
        quote! {
            move |
                mut #mailbox: ::btrun::Mailbox<#msg_enum>,
                #actor_id: ::btrun::model::ActorId,
                #runtime: &'static ::btrun::Runtime
            | async move {
                use ::btlib::bterr;
                use ::btrun::{
                    log,
                    model::{
                        Envelope, ControlMsg, Named, TransResult, ActorError, ActorErrorPayload,
                        TransKind,
                    }
                };
                let #actor_name = #runtime . actor_name(#actor_id);
                let mut state = #state_enum_ident :: #init_state (#init_state_var);
                while let Some(envelope) = #mailbox.recv().await {
                    let new_state = match envelope {
                        Envelope::Call { #msg, #from, mut #reply, .. } => #call_transitions
                        Envelope::Send { #msg, #from, .. } => #send_transitions
                        Envelope::Control(#msg) => #control_transitions
                    };
                    state = new_state;
                    if let #state_enum_ident::#end_ident(_) = &state {
                        break;
                    }
                }
                Ok(#actor_id)
            }
        }
    }

    fn generate_call_transitions(&self, actor: &ActorModel) -> TokenStream {
        self.generate_transitions(actor, |msg_info| msg_info.is_call())
    }

    fn generate_send_transitions(&self, actor: &ActorModel) -> TokenStream {
        self.generate_transitions(actor, |msg_info| !msg_info.is_call())
    }

    fn generate_transitions(
        &self,
        actor: &ActorModel,
        filter: impl Fn(&MsgInfo) -> bool,
    ) -> TokenStream {
        let state_enum_ident = actor.state_enum_ident();
        let msg_enum_ident = self.msg_enum_ident();
        let init_state_type_param = actor.init_state().type_param();
        let msg_enum_kinds = self.msg_enum_kinds_ident();
        let actor_id = self.actor_id_param();
        let transitions = actor.states().values().flat_map(|state| {
            state.methods().values().filter(|method| {
                if let Some(in_msg) = method.def().in_msg() {
                    let msg_info = self.msg_lookup().lookup(in_msg);
                    filter(msg_info)
                } else {
                    false
                }
            })
            .map(|method| {
                let state_type_param = state.type_param();
                let mut output_iter = method.outputs().iter();
                let next_state = output_iter.next()
                    .unwrap_or_else(|| panic!("There are no outputs for method {} in state {}.", method.name(), state.name()));
                let next_state_name = if let OutputKind::State { def, .. } = next_state.kind() {
                    def.state_trait.as_ref()
                } else {
                    panic!("First output of {} method was not a state.", method.name());
                };
                let next_state_var = next_state.var_name();
                let out_states = output_iter
                    .flat_map(|output| {
                        if let OutputKind::State { def, .. } = output.kind() {
                            Some((output.var_name(), def))
                        } else {
                            None
                        }
                    })
                    .map(|(var_name, def)| {
                        let spawning_actor = self.actor_lookup().actor_with_state(&def.state_trait);
                        let spawning_model = self.actors().get(spawning_actor)
                            .unwrap_or_else(|| panic!("There was no actor named {spawning_actor}."));
                        let spawn_function = spawning_model.spawn_function_ident()
                            .unwrap_or_else(|| panic!("Actor {spawning_actor} of kind {:?} has no spawn function.", spawning_model.kind()));
                        let from = self.from_ident();
                        let runtime = self.runtime_param();
                        let method_name = method.name();
                        quote! {
                            if let Err(err) = #spawn_function(#runtime, #from, #var_name).await {
                                log::error!(
                                    "Failed to spawn {} actor after the {} method: {err}",
                                    stringify!(#spawning_actor),
                                    stringify!(#method_name)
                                )
                            }
                        }
                    });
                let out_msgs = method.outputs().iter()
                    .flat_map(|output| {
                        if let OutputKind::Msg { def, .. } = output.kind() {
                            Some((output.var_name(), def))
                        } else {
                            None
                        }
                    })
                    .map(|(var_name, dest)| {
                        if dest.msg.is_reply() {
                            let reply = self.reply_ident();
                            let msg_type = &dest.msg.msg_type;
                            let reply_variant = self.msg_lookup().lookup(&dest.msg).msg_name();
                            let error_msg = format!("Failed to send {} reply.", msg_type);
                            quote! {
                                if let Some(mut reply) = #reply.take() {
                                    if let Err(_) = reply.send(#msg_enum_ident :: #reply_variant (#var_name)) {
                                        return Err(ActorError::new(
                                            bterr!(#error_msg),
                                            ActorErrorPayload {
                                                actor_id: #actor_id,
                                                actor_impl: #init_state_type_param :: actor_impl(),
                                                state: #state_type_param :: state_name(),
                                                message: #msg_enum_kinds :: #msg_type .name(),
                                                kind: TransKind::Receive,
                                            }
                                        ));
                                    }
                                } else {
                                    log::error!(
                                        "Reply to {} message has already been sent.",
                                        #msg_enum_kinds :: #msg_type .name()
                                    );
                                }
                            }
                        } else {
                            todo!("Send message to an owned state or to a service.");
                        }
                    });
                let method_name = method.name();
                let state_name = state.name();
                let msg_name = method.msg_received_input().unwrap().msg_name();
                let out_vars = method.outputs().iter().map(|output| output.var_name());
                quote! {
                    (#state_enum_ident :: #state_name(state), #msg_enum_ident :: #msg_name(msg)) => {
                        match state.#method_name(msg).await {
                            TransResult::Ok(( #( #out_vars ),* )) => {
                                #( #out_states )*
                                #( #out_msgs )*
                                #state_enum_ident :: #next_state_name (#next_state_var)
                            }
                            TransResult::Abort { from, err, .. } => {
                                log::warn!(
                                    "Aborted transition from the {} state while handling the {} message: {}",
                                    stringify!(#state_name),
                                    stringify!(#msg_name),
                                    err
                                );
                                #state_enum_ident :: #state_name(from)
                            }
                            TransResult::Fatal { err, .. } => {
                                return Err(ActorError::new(
                                    err,
                                    ActorErrorPayload {
                                        actor_id: #actor_id,
                                        actor_impl: #init_state_type_param :: actor_impl(),
                                        state: #state_type_param :: state_name(),
                                        message: #msg_enum_kinds :: #msg_name . name(),
                                        kind: TransKind::Receive
                                    }
                                ));
                            }
                        }
                    },
                }
            })
        });
        quote! {
            match (state, msg) {
                #( #transitions )*
                (state, msg) => {
                    log::error!(
                        "Unexpected message {} in state {}.", msg.name(), state.name()
                    );
                    state
                }
            }
        }
    }

    fn generate_control_transitions(&self, _actor: &ActorModel) -> TokenStream {
        quote! {
            todo!()
        }
    }

    fn generate_state_enum(&self, actor: &ActorModel) -> TokenStream {
        let enum_ident = actor.state_enum_ident();
        let end_ident = self.end_ident();
        let pairs: Vec<_> = actor
            .states()
            .values()
            .map(|state| {
                let trait_type = state.name();
                let type_param = format_ident!("T{trait_type}");
                (trait_type, Some(type_param))
            })
            .chain(std::iter::once((end_ident, None)))
            .collect();
        let type_param_decls = pairs
            .iter()
            .filter(|(_, type_param)| type_param.is_some())
            .map(|(trait_type, type_param)| quote! { #type_param: #trait_type });
        let decl_type_params = quote! { < #( #type_param_decls ),* > };
        let variants = pairs
            .iter()
            .filter(|(_, type_param)| type_param.is_some())
            .map(|(trait_type, type_param)| quote! { #trait_type(#type_param) })
            .chain(std::iter::once(
                quote! { #end_ident(::btrun::model::#end_ident) },
            ));
        let type_params = pairs.iter().flat_map(|(_, type_param)| type_param);
        let name_pairs: Vec<_> = pairs
            .iter()
            .map(|(trait_type, _)| {
                let trait_type_string = trait_type.to_string();
                let name_ident = format_ident!("{}_NAME", trait_type_string.pascal_to_snake().to_uppercase());
                let name_decl = quote! {
                    static #name_ident: ::btrun::model::Lazy<::std::sync::Arc<String>> =
                        ::btrun::model::Lazy::new(|| ::std::sync::Arc::new(#trait_type_string.into()));
                };
                let match_branch = quote! {
                    Self::#trait_type(_) => #name_ident.clone(),
                };
                (name_decl, match_branch)
            })
            .collect();
        let name_decls = name_pairs.iter().map(|(name_decl, _)| name_decl);
        let name_match_branches = name_pairs.iter().map(|(_, match_branch)| match_branch);
        quote! {
            enum #enum_ident #decl_type_params {
                #( #variants ),*
            }

            impl #decl_type_params ::btrun::model::Named for #enum_ident<#( #type_params ),*> {
                fn name(&self) -> ::std::sync::Arc<String> {
                    #( #name_decls )*
                    match self {
                        #( #name_match_branches )*
                    }
                }
            }
        }
    }
}

impl MethodModel {
    /// Generates the tokens for the code which implements this transition.
    fn generate_tokens(&self) -> TokenStream {
        let method_ident = self.name().as_ref();
        let msg_args = self.inputs().iter();
        let output_decls = self.outputs().iter().flat_map(|output| output.decl());
        let output_types = self.outputs().iter().flat_map(|output| output.type_name());
        let future_name = self.future();
        quote! {
            #( #output_decls )*
            type #future_name: Send + ::std::future::Future<
                Output = ::btrun::model::TransResult<Self, ( #( #output_types ),* )>
            >;
            fn #method_ident(self #( , #msg_args )*) -> Self::#future_name;
        }
    }
}
